/*-------------------------------------------------------------------------+
 Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static org.fortiss.af3.exploration.alg.util.ExplorationFeatureUtils.getTransformedOutputsByFeatures;
import static org.fortiss.af3.exploration.model.solutions.SolutionState.OPTIMIZED;
import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createExplorationResult;
import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createExplorationSolution;

import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.ProvidingPhenotype;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.internal.transformation.ExplorationTransformationService;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.util.ExplorationModelElementFactory;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.service.ITransformationService;
import org.opt4j.core.Individual;
import org.opt4j.core.Objective;
import org.opt4j.core.Objectives;
import org.opt4j.core.optimizer.Archive;

import com.google.inject.TypeLiteral;

/**
 * Converts the results of the internal Opt4J representation of solutions into an
 * {@link ExplorationSolution}.
 * 
 * @author barner
 * @author diewald
 */
public class Opt4JExplorationSolutionTransformer {

	/** Constructor. */
	private Opt4JExplorationSolutionTransformer() {
		// Prevent instantiation. Stateless.
	}

	/**
	 * Transforms the solution set given as an {@link Archive} produced by Opt4J into an
	 * {@link ExplorationSolution} of the EMM.
	 * 
	 * @param archive
	 *            {@link Archive} produced by Opt4J.
	 * @param targetObjectiveAssocs
	 *            associative map of EMM {@link ExplorationTarget}s from the converted input
	 *            {@link ExplorationSpecification} to Opt4J-internal {@link Objective}s.
	 * @param expSpec
	 *            specification that contains the {@link ExplorationTarget}s input to the Opt4J
	 *            backend.
	 * @param enabledFeatures
	 *            {@link IExplorationFeature}s enabled for this DSE run.
	 * @return corresponding {@link ExplorationSolution} of the EMM.
	 * @throws ExplorationException
	 *             if inconsistencies between the in- and output requirements are found.
	 */
	public static ExplorationSolution transform(Archive archive,
			Map<ExplorationTarget<?>, Objective> targetObjectiveAssocs,
			ExplorationSpecification expSpec, Collection<IExplorationFeature> enabledFeatures)
			throws ExplorationException {
		ExplorationSolution explorationSolution = createExplorationSolution();

		int solutionIdx = 0;
		for(Individual individual : archive) {
			SingleExplorationSolution solution = createSingleExplorationSolution(
					individual.getObjectives(), targetObjectiveAssocs);
			explorationSolution.getSolutions().add(solution);
			// TODO: Not type-safe. Opt4J has to be modified here.
			addOutputSuperSets(solution, ((ProvidingPhenotype)individual.getPhenotype()),
					enabledFeatures);
			solution.setSolutionState(OPTIMIZED);
			solution.setName("Solution " + solutionIdx++);
		}
		explorationSolution.setExplorationSpec(expSpec);

		return explorationSolution;
	}

	/** Transforms {@link Objectives} into a {@link SingleExplorationSolution}. */
	private static <T> SingleExplorationSolution createSingleExplorationSolution(
			Objectives objectives, Map<ExplorationTarget<?>, Objective> targetObjectiveAssocs)
			throws ExplorationException {
		SingleExplorationSolution solution =
				ExplorationModelElementFactory.createSingleExplorationSolution();
		for(Entry<ExplorationTarget<?>, Objective> targetAssoc : targetObjectiveAssocs.entrySet()) {
			// Determine the ExplorationTarget class to identify the correct return value type.
			Objective objective = targetAssoc.getValue();
			@SuppressWarnings("unchecked") ExplorationTarget<T> explorationTarget =
					(ExplorationTarget<T>)targetAssoc.getKey();
			Class<?> expectedResultType = explorationTarget.getResultType();
			if(expectedResultType == null) {
				throw new ExplorationException(
						"Could not determine the result type of the exploration target " +
								explorationTarget.getName() +
								". The ExplorationTarget was wrongly constructed.");
			}

			// Type check the result vs. the expected type defined by the exploration target.
			Object result = objectives.get(objective).getValue();
			if(result != null && !result.getClass().equals(expectedResultType)) {
				// Note: if the result is null, a serious constraint violation is indicated. Its
				// corresponding solution will be filtered out in a later stage.
				throw new IllegalArgumentException(
						"Incompatible evaluation result type for the exploration target " +
								explorationTarget.getName() + ".");
			}
			@SuppressWarnings("unchecked") T castedResult = (T)result;

			ExplorationResult<T> explorationResult = createExplorationResult();
			explorationResult.setResult(castedResult);
			solution.put(explorationTarget, explorationResult);
		}

		return solution;
	}

	/**
	 * Adds the producable {@link SuperSet}s as a {@link SuperSetMap} to the given
	 * {@link SingleExplorationSolution}. It delegates the transformation operation to the
	 * {@link ITransformationService}.
	 * 
	 * @param solution
	 *            where the transformed {@link SuperSet}s are added.
	 * @param phenotype
	 *            output solution produced by the DSE.
	 * @param enabledFeatures
	 *            Features enabled for this DSE run.
	 * @return Constructed {@link SuperSetMap} from the exploration solutions.
	 * @throws ExplorationException
	 *             if the transformation service fails.
	 */
	@SuppressWarnings("unchecked")
	private static SuperSetMap addOutputSuperSets(SingleExplorationSolution solution,
			ProvidingPhenotype phenotype, Collection<IExplorationFeature> enabledFeatures)
			throws ExplorationException {
		Set<TypeLiteral<SuperSet<? extends IModelElement>>> outputSets =
				getTransformedOutputsByFeatures(enabledFeatures);
		ExplorationTransformationService eTS = ExplorationTransformationService.getInstance();
		SuperSetMap solutionSets = eTS.transform(enabledFeatures, phenotype, outputSets);
		// Add the transformed SuperSets to the passed ExplorationSolution.
		solutionSets.keySet().forEach(k -> solution.putSolutionModel((Class<IModelElement>)k,
				solutionSets.get((Class<IModelElement>)k)));
		return solutionSets;
	}
}
