/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext;

import static java.util.Arrays.asList;

import org.apache.commons.collections4.IterableUtils;
import org.fortiss.af3.exploration.alg.dse.RandomExploration;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.genotype.ProvidingGenotype;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.operators.crossover.CrossoverGenericImplementation;
import org.opt4j.operators.crossover.Pair;

import com.google.inject.Singleton;

/**
 * Implementation of the {@link CrossoverGenericImplementation} for exploration processes: Only a
 * single crossover operation is applied in contrast to applying the mutate operation to each sub
 * {@link Genotype}. The latter wastes computations since the genotypes are modified but overridden
 * in the decode phase to maintain consistency.
 * 
 * @author diewald
 */
@Singleton
public class CrossoverGenericProcessImplementation extends CrossoverGenericImplementation {

	/** References the random number generator instance. */
	private RandomExploration randExp = RandomExploration.getInstance();

	/**
	 * Forwards a randomly selected {@link Genotype} contained in the given
	 * {@link CompositeGenotype}s to an operator that performs the crossover operation on them.
	 */
	@Override
	protected Pair<Genotype> crossoverComposite(CompositeGenotype<?, ?> p1,
			CompositeGenotype<?, ?> p2) {
		CompositeGenotype<Object, Genotype> o1 = p1.newInstance();
		CompositeGenotype<Object, Genotype> o2 = p2.newInstance();

		int selectedGenotypeIdx = randExp.nextInt(p1.keySet().size());
		// No NPE checked for the getter --> would indicate a framework problem.
		final Object key = IterableUtils.get(p1, selectedGenotypeIdx).getKey();

		final Genotype g1 = p1.get(key);
		final Genotype g2 = p2.get(key);
		Pair<Genotype> genotype = crossover(g1, g2);
		o1.put(key, genotype.getFirst());
		o2.put(key, genotype.getSecond());

		if(g1 instanceof ProvidingGenotype && g2 instanceof ProvidingGenotype) {
			o1.setModifiedGenotypes(asList(g1));
			o2.setModifiedGenotypes(asList(g2));
		}

		Pair<Genotype> offspring = new Pair<Genotype>(o1, o2);
		return offspring;
	}
}
