/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext;

import static com.google.common.collect.Maps.newHashMap;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;

import com.google.inject.Provides;
import com.google.inject.TypeLiteral;

import net.bytebuddy.ByteBuddy;
import net.bytebuddy.description.annotation.AnnotationDescription;
import net.bytebuddy.description.modifier.Visibility;
import net.bytebuddy.dynamic.DynamicType.Builder.FieldDefinition.Optional.Valuable;
import net.bytebuddy.dynamic.DynamicType.Builder.MethodDefinition;
import net.bytebuddy.dynamic.DynamicType.Builder.MethodDefinition.ReceiverTypeDefinition;
import net.bytebuddy.implementation.FieldAccessor;
import net.bytebuddy.implementation.MethodCall;

/**
 * Generates a EncodingProviderModule holding all generated Providers (prototypes). First,
 * the Provider classes get constructed and they are initialized as prototypes such that
 * the @Provides
 * methods are available for binding. At this stage, these providers contain no
 * encoding instances such that provisioning would fail.
 * In order to acquire these instance through the generated providers, the
 * {@link #generateInstanceProviders(Iterable)} must be called that
 * creates new Providers holding these instances. For this operation, the prototype module must be
 * available.
 * 
 * @author diewald
 */
public abstract class EncodingProviderGenerator<T> {

	/** Name of the field(s) holding the encoding. */
	protected String fieldName = "encoding";

	/** References the internal prototype provider for concrete encoding providers. */
	protected EncodingProviderModule<T> protoModule;

	/**
	 * Association of optional {@link Annotation}s of {@link Provides} methods per returned type.
	 */
	private Map<TypeLiteral<? extends T>, Class<? extends Annotation>> providerAnnotations =
			new HashMap<>();

	/** Constructor. */
	protected EncodingProviderGenerator(Collection<TypeLiteral<? extends T>> encodingTypesToProvide,
			Map<TypeLiteral<? extends T>, Class<? extends Annotation>> providerAnnotations) {
		this.providerAnnotations.putAll(providerAnnotations);
		this.protoModule = generatePrototypes(encodingTypesToProvide);
	}

	/**
	 * Returns the {@link EncodingProviderModule} that is setup during construction time and
	 * contains the class objects to construct concrete encoding providers.
	 */
	public EncodingProviderModule<T> getEncodingProviderModule() {
		return protoModule;
	}

	/**
	 * Generates a EncodingProviderModule holding all generated prototype Providers. See
	 * {@link EncodingProviderModule} for details.
	 * 
	 * @return EncodingProviderModule holding all generated providers.
	 */
	private EncodingProviderModule<T>
			generatePrototypes(Collection<TypeLiteral<? extends T>> encodingTypesToProvide) {
		Map<TypeLiteral<? extends T>, EncodingProvider<? extends T>> providersMap = newHashMap();
		for(TypeLiteral<? extends T> elementKey : encodingTypesToProvide) {
			providersMap.put(elementKey, createProvider(elementKey));
		}
		return new EncodingProviderModule<T>(providersMap);
	}

	/**
	 * Generates a EncodingProviderModule holding all generated Providers. In this step, also the
	 * Providers themselves get created. A pre-constructed prototype {@link EncodingProviderModule}
	 * is required for this operation.
	 * 
	 * @return EncodingProviderModule holding all generated providers.
	 */
	@SuppressWarnings("unchecked")
	public synchronized EncodingProviderModule<T>
			generateInstanceProviders(Iterable<Entry<TypeLiteral<? extends T>, T>> encodings) {
		Map<TypeLiteral<? extends T>, EncodingProvider<? extends T>> providersMap = new HashMap<>();
		for(Entry<TypeLiteral<? extends T>, T> entry : encodings) {
			EncodingProvider<? extends T> protoProvider =
					protoModule.getEncodingProviderOf(entry.getKey());
			EncodingProvider<T> instanceProvider =
					createInstanceOf((Class<? extends EncodingProvider<T>>)protoProvider.getClass(),
							entry.getKey(), entry.getValue());
			providersMap.put(entry.getKey(), instanceProvider);
		}
		return new EncodingProviderModule<T>(providersMap);
	}

	/** Creates a provider for the given encoding. */
	@SuppressWarnings("rawtypes")
	protected EncodingProvider<T> createProvider(TypeLiteral<? extends T> encodingKey) {
		Valuable<EncodingProvider> subClassField = subClassWithField(encodingKey);
		ReceiverTypeDefinition<EncodingProvider> withAssigningConstructor =
				createAssigningConstructor(subClassField, encodingKey);
		MethodDefinition<EncodingProvider> finishedProvider =
				createProvidesMethod(withAssigningConstructor, encodingKey);

		// @CodeFormatterOff
		@SuppressWarnings("unchecked")
		Class<? extends EncodingProvider<T>> providerClass = (Class<? extends EncodingProvider<T>>)
				finishedProvider
				.make()
				.load(getClass().getClassLoader())
				.getLoaded();
		// @CodeFormatterOn

		return createPrototypeOf(providerClass, encodingKey);
	}

	/**
	 * Creates a subclass of the {@link EncodingProvider} and adds a field to hold the given
	 * {@code encoding}.
	 */
	@SuppressWarnings("rawtypes")
	protected Valuable<EncodingProvider> subClassWithField(TypeLiteral<? extends T> encodingType) {
		return new ByteBuddy().subclass(EncodingProvider.class)
				.name("Provides" + constructProviderName(encodingType))
				.defineField(fieldName, encodingType.getType(), Visibility.PRIVATE);
	}

	/**
	 * Creates a constructor that assigns the given input encoding to a {@code previous}ly
	 * constructed field.
	 */
	@SuppressWarnings("rawtypes")
	protected ReceiverTypeDefinition<EncodingProvider> createAssigningConstructor(
			Valuable<EncodingProvider> previous, TypeLiteral<? extends T> encodingKey) {
		try {
			// @CodeFormatterOff
			return previous
					.defineConstructor(Visibility.PUBLIC)
					.withParameter(encodingKey.getType())
					.intercept(MethodCall
							.invoke(EncodingProvider.class.getDeclaredConstructor())
							.onSuper()
							.andThen(FieldAccessor.ofField(fieldName).setsArgumentAt(0)));
			// @CodeFormatterOn
		} catch(NoSuchMethodException | SecurityException e) {
			throw new ExplorationRuntimeException(
					"Could not generate the setter constructor for the generated" +
							" provider of the encoding " +
							encodingKey.getRawType().getSimpleName() + ".");
		}
	}

	/**
	 * Creates a getter method for the field named {@link #fieldName} and annotates it with
	 * 
	 * @Provides.
	 */
	@SuppressWarnings("rawtypes")
	protected MethodDefinition<EncodingProvider> createProvidesMethod(
			ReceiverTypeDefinition<EncodingProvider> withConstructor,
			TypeLiteral<? extends T> encodingKey) {
		// @CodeFormatterOff
		MethodDefinition<EncodingProvider> providesMethod = withConstructor
				.defineMethod("get"+constructProviderName(encodingKey), encodingKey.getType(), Visibility.PUBLIC)
				.intercept(FieldAccessor.ofField(fieldName))
				.annotateMethod(AnnotationDescription.Builder.ofType(Provides.class).build());
		// @CodeFormatterOn

		Class<? extends Annotation> providerAnnotation = providerAnnotations.get(encodingKey);
		if(providerAnnotation != null) {
			providesMethod = providesMethod.annotateMethod(
					AnnotationDescription.Builder.ofType(providerAnnotation).build());
		}

		return providesMethod;
	}

	/**
	 * Creates an instance of the given {@link EncodingProvider} class that wraps the given
	 * {@code encoding}.
	 */
	protected <P extends EncodingProvider<T>> P createInstanceOf(Class<P> providerClass,
			TypeLiteral<? extends T> encodingKey, T encoding) {
		try {
			Constructor<P> providerConstructor =
					providerClass.getConstructor(encodingKey.getRawType());
			// Debug code: To see if the getters are created correctly.
			// for(Method method : providerClass.getDeclaredMethods()) {
			// System.out.println(method.getName());
			// }
			P providerInst = providerConstructor.newInstance(encoding);
			return providerInst;
		} catch(NoSuchMethodException | SecurityException | InstantiationException |
				IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			throw new ExplorationRuntimeException(
					"Could not create an instance of the generated Provider for the Phenotype " +
							encoding.getClass().getSimpleName() + ".",
					e);
		}
	}

	/**
	 * Creates an instance of the given {@link EncodingProvider} class that wraps the given
	 * {@code encoding}.
	 */
	protected EncodingProvider<T> createPrototypeOf(
			Class<? extends EncodingProvider<T>> providerClass,
			TypeLiteral<? extends T> encodingKey) {
		try {
			Constructor<? extends EncodingProvider<T>> providerConstructor =
					providerClass.getConstructor();
			// Debug code: To see if the getters are created correctly.
			// for(Method method : providerClass.getDeclaredMethods()) {
			// System.out.println(method.getName());
			// }
			EncodingProvider<T> providerInst = providerConstructor.newInstance();
			return providerInst;
		} catch(NoSuchMethodException | SecurityException | InstantiationException |
				IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			throw new ExplorationRuntimeException(
					"Could not create an instance of the generated Provider for the Phenotype " +
							encodingKey.getType() + ".",
					e);
		}
	}

	/** Constructs a unique name for the given encoding {@link TypeLiteral}. */
	private String constructProviderName(TypeLiteral<? extends T> encodingKey) {
		String name = encodingKey.getRawType().getSimpleName() + "_";
		Type type = encodingKey.getType();
		if(type instanceof ParameterizedType) {
			for(Type genType : ((ParameterizedType)type).getActualTypeArguments()) {
				name = name + genType.getTypeName();
			}
		}
		name = name.replaceAll("[.]", "_");
		name = name.replaceAll("[^a-zA-Z_]", "");
		return name;
	}
}
