/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.create;

import static java.util.stream.Collectors.toMap;

import java.lang.annotation.Annotation;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.DseInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.InfeasibleEncodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.constraint.ConstrainedArtifactMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.constraint.IConstrainedArtifactMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.DecodeInterceptor;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.GuiceDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.PhenotypeCache;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.genotype.GenotypeProviderGenerator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.genotype.ProvidingGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.module.CreatorModule;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Creator;

import com.google.inject.AbstractModule;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.ProvisionException;
import com.google.inject.TypeLiteral;

/**
 * {@link CreatorModule} that is the entry point and majestro for the create phase of the DI-based
 * DSE process exploration.
 * 
 * @author diewald
 */
public class GuiceCreator extends AbstractModule implements Creator<ProvidingGenotype> {

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<CreatorModule<?>> enabledCreators = new HashSet<>();

	/** Set of {@link CreatorModule}s provided by the enabled exploration features. */
	private Collection<Class<? extends CreatorModule<?>>> enabledCreatorTypes;

	/** Set of {@link DecoderModule} types required as dependencies by {@link CreatorModule}s . */
	private Collection<TypeLiteral<? extends DecoderModule<?>>> enabledDecoderTypes;

	/** References the top-most injector of the DSE. */
	private Injector dseInjector;

	/** Holds the {@link CreatorModule}s. */
	protected Injector createInjector;

	/** Constructor. Remembers the demanded {@link Phenotype}ss to be produced. */
	@Inject
	public GuiceCreator(DseInjectorProvider dseInjectorProvider) {
		this.dseInjector = dseInjectorProvider.getInjector();
	}

	/**
	 * Declares the set of {@link CreatorModule}s required to produce the needed Genotypes as
	 * defined by the input exploration specifications
	 */
	public void
			setEnabledCreators(Collection<Class<? extends CreatorModule<?>>> enabledCreatorTypes) {
		this.enabledCreatorTypes = enabledCreatorTypes;
	}

	/** Sets the decoders that are required by {@link CreatorModule}s. */
	public void setEnabledDecoders(
			Collection<TypeLiteral<? extends DecoderModule<?>>> enabledDecoderTypes) {
		this.enabledDecoderTypes = enabledDecoderTypes;
	}

	/** {@inheritDoc} */
	@Override
	public ProvidingGenotype create() {
		boolean isInfeasible = false;

		// We do not need to sync the block below as Creators are not parallelized.
		CreateInterceptor cInterceptor = null;
		if(createInjector == null) {
			createInjector = dseInjector.createChildInjector(this);
		}

		// TODO(AD): Use binding / injection to create the cMap.
		ExplorationSpecification expSpec =
				createInjector.getInstance(ExplorationSpecification.class);
		IConstrainedArtifactMap cMap = new ConstrainedArtifactMap(expSpec);
		cInterceptor = createInjector.getInstance(CreateInterceptor.class);
		cInterceptor.setInterceptParameters(createInjector, cMap, Genotype.class,
				GenotypeCache.class);
		// Clean the cache for each new run.
		GenotypeCache gCache = createInjector.getInstance(GenotypeCache.class);
		gCache.clear();
		PhenotypeCache pCache = createInjector.getInstance(PhenotypeCache.class);
		pCache.clear();

		GenotypeCache genotypeCache = createInjector.getInstance(GenotypeCache.class);
		genotypeCache.clear();

		for(CreatorModule<?> creator : enabledCreators) {
			try {
				DecodeInterceptor interceptor = createInjector.getInstance(DecodeInterceptor.class);
				interceptor.setInterceptParameters(createInjector, cMap, Phenotype.class,
						PhenotypeCache.class);

				createInjector.getInstance(Key.get(creator.getProvidedType(), Genotyped.class));
			} catch(ProvisionException e) {
				if(e.getCause() instanceof InfeasibleEncodingException) {
					isInfeasible = true;
				} else {
					throw e;
				}
			}
		}

		return createProvidingGenotype(genotypeCache, cMap, isInfeasible);
	}

	/**
	 * Constructs a {@link ProvidingGenotype} from the created {@link Genotype}s to be consumed by a
	 * {@link GuiceDecoder}-
	 */
	private ProvidingGenotype createProvidingGenotype(GenotypeCache genotypeCache,
			IConstrainedArtifactMap cMap, boolean isInfeasible) {
		Map<TypeLiteral<? extends Genotype>, Class<? extends Annotation>> providerAnnotations =
				genotypeCache.keySet().stream().collect(toMap(g -> g, g -> Genotyped.class));
		GenotypeProviderGenerator generator =
				new GenotypeProviderGenerator(genotypeCache.keySet(), providerAnnotations);
		// With a composite genotype of composite genotypes and a recursive collect, we could also
		// support arbitrarily deep nested structure, but this not needed atm.
		CompositeGenotype<TypeLiteral<? extends Genotype>, Genotype> compGenotype =
				new CompositeGenotype<>(genotypeCache);
		ProvidingGenotype providingGenotype = new ProvidingGenotype(compGenotype, generator, cMap);
		if(isInfeasible) {
			providingGenotype.setInfeasible();
		}
		return providingGenotype;
	}

	/** {@inheritDoc} */
	@Override
	protected void configure() {
		install(new GenotypeCache());
		install(PhenotypeCache.getPrototype());

		for(Class<? extends CreatorModule<?>> moduleType : enabledCreatorTypes) {
			CreatorModule<?> module = dseInjector.getInstance(moduleType);
			enabledCreators.add(module);
			install(module);
		}

		for(TypeLiteral<? extends DecoderModule<?>> moduleType : enabledDecoderTypes) {
			DecoderModule<?> module = dseInjector.getInstance(Key.get(moduleType));
			install(module);
		}
	}
}
