/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.graph;

import static com.google.common.collect.Sets.newHashSet;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.module.DecoderModule;
import org.opt4j.core.problem.Decoder;

import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.TypeLiteral;

/**
 * Finds the set of required {@link Decoder}s that is needed to construct a given set
 * of {@link Phenotype}s. Therefore, a visitor is constructed that remembers the bindings to
 * {@link Decoder}s providing {@link Phenotype} objects. The visitor traverses the decoding
 * {@link Injector} to capture the dependencies of the decoding phase. Here, all potentially used
 * {@link Decoder}s have to be analyzed, i.e. all {@link Decoder}s shipped by the set of enabled
 * {@link IExplorationFeature}s must be bound in the given {@link Injector}.
 * This class is inspired by the InjectorGrapher of the Guice Grapher extension.
 * 
 * @author diewald
 */
public class DecoderFinder
		extends BindingWalker<TypeMemorizingDependencyVisitor<DecoderModule<?>>> {

	/** Constructor. */
	public DecoderFinder() {
		super(new TypeMemorizingDependencyVisitor<>(newHashSet(Decoder.class)));
	}

	/**
	 * Collects the set of {@link Decoder}s that are required to decode the set of given
	 * {@link Phenotype}s.
	 * 
	 * @param encodings
	 *            Set of requested artifacts (e.g., a schedule).
	 * @param injector
	 *            DSE-wide injector to analyze the {@link Decoder} dependencies.
	 * @return Set of required {@link Decoder}s.
	 */
	public Set<TypeLiteral<? extends DecoderModule<?>>> getDecodersFor(
			Map<TypeLiteral<?>, Class<? extends Annotation>> encodings, Injector injector) {
		Set<Key<?>> phenotypeKeys = new HashSet<>();
		for(Entry<TypeLiteral<?>, Class<? extends Annotation>> entry : encodings.entrySet()) {
			if(entry.getValue() != null) {
				phenotypeKeys.add(Key.get(entry.getKey().getType(), entry.getValue()));
			} else {
				phenotypeKeys.add(Key.get(entry.getKey().getType()));
			}
		}
		iterateBindings(injector, phenotypeKeys);
		return keyVisitor.getFoundSubTypes();
	}
}
