/*
 * Copyright (C) 2008 Google Inc.
 * Copyright (C) 2018 fortiss GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.graph;

import java.util.Collection;

import com.google.common.collect.ImmutableSet;
import com.google.inject.Binder;
import com.google.inject.Binding;
import com.google.inject.Key;
import com.google.inject.spi.ConstructorBinding;
import com.google.inject.spi.ConvertedConstantBinding;
import com.google.inject.spi.DefaultBindingTargetVisitor;
import com.google.inject.spi.HasDependencies;
import com.google.inject.spi.InstanceBinding;
import com.google.inject.spi.LinkedKeyBinding;
import com.google.inject.spi.ProviderBinding;
import com.google.inject.spi.ProviderInstanceBinding;
import com.google.inject.spi.ProviderKeyBinding;

/**
 * Base visitor to scan for artifacts provided by some provides method and for the classes
 * that provide artifacts (@Provides Methods). Each relevant visit method is dispatched to the
 * abstract {@link #visitHasDependencies(HasDependencies, Binding)} method.
 * This class is a modified version of the TransitiveDependencyVisitor class of the Grapher
 * extension of google Guice.
 *
 * @author phopkins@gmail.com (Pete Hopkins)
 * @author diewald@fortiss.org (Alexander Diewald)
 */
public abstract class DependencyVisitorBase
		extends DefaultBindingTargetVisitor<Object, Collection<Key<?>>> {

	/** Iterates over the dependency tree spanned by the {@link Binder}. */
	protected abstract Collection<Key<?>> visitHasDependencies(HasDependencies hasDependencies,
			Binding<?> binding);

	/** Visits a {@link ConstructorBinding}. */
	@Override
	public Collection<Key<?>> visit(ConstructorBinding<?> binding) {
		return visitHasDependencies(binding, binding);
	}

	/** Visits a {@link ConvertedConstantBinding}. */
	@Override
	public Collection<Key<?>> visit(ConvertedConstantBinding<?> binding) {
		return visitHasDependencies(binding, binding);
	}

	/** Visits a {@link InstanceBinding}. */
	@Override
	public Collection<Key<?>> visit(InstanceBinding<?> binding) {
		return visitHasDependencies(binding, binding);
	}

	/** Visits a {@link LinkedKeyBinding}. */
	@Override
	public Collection<Key<?>> visit(LinkedKeyBinding<?> binding) {
		return ImmutableSet.<Key<?>>of(binding.getLinkedKey());
	}

	/** Visits a {@link ProviderBinding}. */
	@Override
	public Collection<Key<?>> visit(ProviderBinding<?> binding) {
		return ImmutableSet.<Key<?>>of(binding.getProvidedKey());
	}

	/** Visits a {@link ProviderInstanceBinding}. */
	@Override
	public Collection<Key<?>> visit(ProviderInstanceBinding<?> binding) {
		return visitHasDependencies(binding, binding);
	}

	/** Visits a {@link ProviderKeyBinding}. */
	@Override
	public Collection<Key<?>> visit(ProviderKeyBinding<?> binding) {
		return ImmutableSet.<Key<?>>of(binding.getProviderKey());
	}

	/** @since 4.0 */
	@Override
	public Collection<Key<?>> visitOther(Binding<?> binding) {
		return ImmutableSet.of();
	}
}
