/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getAnnotationChecked;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.platform.hierarchic.model.annotation.TransmissionUnitPower;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.timing.model.annotation.TransmissionUnitBandwidth;

/**
 * Adapter between the AF3 {@link TransmissionUnit} and the DSE.
 * 
 * @author barner
 */
public class AF3TransmissionUnitAdapter extends AF3ResourceAdapter<TransmissionUnit>
		implements ITransmissionUnitAdapter {
	/** Conversion constant for the message size. */
	private static final long MBYTES_TO_BYTES = 1000 * 1000;

	/** Conversion bytes -> bits. */
	private static final long BYTES_TO_BITS = 8;

	/**
	 * {@link IResourceAdapter}s that are connected with this {@link AF3TransmissionUnitAdapter}.
	 */
	Collection<IResourceAdapter> connectedResources;

	/** {@link TransmissionUnit} bandwidth annotation */
	TransmissionUnitBandwidth transmissionUnitBandwidth;

	/** {@link TransmissionUnit} power consumptions annotation */
	Double transmissionUnitPower;

	/**
	 * Creates an {@link ISignalAdapter} for an {@link TransmissionUnit}.
	 * 
	 * @param transmissionUnit
	 *            present in the target platform and for which to generate an adapter.
	 * @throws InvalidModelException
	 *             if a required annotation cannot be found or if it has an invalid value.
	 */
	public AF3TransmissionUnitAdapter(TransmissionUnit transmissionUnit)
			throws InvalidModelException {
		super(transmissionUnit);
		transmissionUnitPower = getAnnotationChecked(transmissionUnit, TransmissionUnitPower.class)
				.getPowerInWatt();
		transmissionUnitBandwidth =
				getAnnotationChecked(transmissionUnit, TransmissionUnitBandwidth.class);
		if(transmissionUnitBandwidth.getBandwidth_MBitS() == 0.0) {
			throw new InvalidModelException(transmissionUnit,
					"A bandwidth of 0 MB/s is not a valid attribute.");
		}
	}

	/** {@inheritDoc} */
	@Override
	public double getPower(long messageSizeInBits) {
		return messageSizeInBits /
				(MBYTES_TO_BYTES * transmissionUnitBandwidth.getBandwidth_MBitS()) *
				transmissionUnitPower;
	}

	/** {@inheritDoc} */
	@Override
	public double getTransmissionDuration(double messageSizeMB) {
		return (messageSizeMB * BYTES_TO_BITS) / transmissionUnitBandwidth.getBandwidth_MBitS();
	}
}
