/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.Collection;
import java.util.Set;

import org.eclipse.draw2d.graph.DirectedGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.jgrapht.graph.DefaultListenableGraph;
import org.jgrapht.graph.DirectedWeightedMultigraph;

/**
 * Interface for {@link ITaskGraph}s which are the (flattened) DSE-internal representations of a
 * (hierarchical) logical architecture.
 * 
 * @author diewald
 */
public interface ITaskGraph<T extends ITaskAdapter> {
	// TODO: Get rid of this method!
	/** Returns the underlying {@link DefaultListenableGraph}. */
	DirectedWeightedMultigraph<T, SignalAdapterWeightedEdge> getGraph();

	/** Returns the underlying {@link DefaultListenableGraph}. */
	DirectedWeightedMultigraph<ITaskAdapter, SignalAdapterWeightedEdge> getGenericGraph();

	/**
	 * Updates the period of this {@link ITaskGraph}: Determine the common period of the tasks
	 * within
	 * an independent task graph if it exists. Otherwise, set it to 0.
	 * 
	 * @throws InvalidTimingModelException
	 *             if the periods of the {@link ITaskGraph}'s task are inconsistent.
	 */
	void updateHyperPeriod() throws InvalidTimingModelException;

	/** Returns the {@link ITaskGraph}s period. */
	Double getHyperPeriod();

	/** Returns the set of deployable components (=tasks) of this {@link ITaskGraph}. */
	Set<T> getTasks();

	/**
	 * Returns the set of {@link SignalAdapterWeightedEdge}s referring to {@link ISignalAdapter}s.
	 */
	Set<SignalAdapterWeightedEdge> getSignals();

	/**
	 * Checks whether {@code this} {@link ITaskGraph} contains the given {@link ITaskAdapter} or an
	 * instance of it.
	 */
	boolean containsTask(ITaskAdapter task);

	/**
	 * Checks whether {@code this} {@link ITaskGraph} contains any of the given {@link ITaskAdapter}
	 * s.
	 */
	boolean containsAnyTaskOf(Collection<ITaskAdapter> taskSet);

	/**
	 * Adds the given {@link ITaskAdapter} to the {@link ITaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}.
	 */
	boolean addTask(T task);

	/**
	 * Adds the given {@code task} to the {@link ITaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}, and further makes it accessible if the given
	 * {@code refTask} is queried.
	 */
	boolean addTask(T task, T refTask);

	/**
	 * Adds all given tasks to the task graph.
	 * 
	 * @param tasks
	 *            collection of tasks to add to the graph.
	 * @return whether the tasks were successfully added.
	 */
	boolean addAllTask(Collection<T> tasks);

	/** Adds a signal to the task graph that defines a data exchange between two tasks. */
	boolean addSignal(ISignalAdapter signalAdapter);

	/** Adds signals to the task graph that are used to exchange data between tasks. */
	boolean addAllSignals(Collection<ISignalAdapter> signalAdapters);

	/**
	 * Replaces the given {@code task} with the given {@code replacement}. Thereby, all edges are
	 * reconnected to the replacement.
	 */
	ITaskAdapter replaceTask(T task, T replacement);

	/** Removes the given {@code task} and its data dependencies from the {@link ITaskGraph}. */
	void removeTask(ITaskAdapter task);

	/** Creates a Dialog that illustrates this {@link ITaskGraph}. */
	void display();

	/** Creates and returns a copy of {@code this} {@link ITaskGraph}. */
	<S extends ITaskGraph<T>> S copy();

	/** Registers a fault detector for a given task. */
	void addFaultDetector(ITaskAdapter deployableComponent, FaultDetector faultDetector);

	/** Returns the fault detector corresponding to the given Task and ID. */
	// TODO: Remove, bad design. Fault detectors are tasks and should be added in a decoder to
	// another phenotype.
	FaultDetector getFaultDetector(ITaskAdapter task, int id);

	/** Returns a random fault detector index / ID for the given task. */
	int getRandomFaultDetectorId(ITaskAdapter deployableComponent);
}
