/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.partition.evaluator;

import static java.lang.Math.expm1;
import static java.lang.Math.min;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Evaluate;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Validates;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.partition.variable.Partition;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.common.evaluator.ConstraintEvaluatorBase;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.moea.model.constraints.SameSilInPartitionConstraint;
import org.fortiss.af3.safety.model.SIL;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.inject.Provides;

/**
 * Evaluates whether all {@link ITaskAdapter}s mapped to {@link Partition}s have the same
 * SIL/ASIL/DAL value and calculates a penality based on the difference in the Integrity Levels of
 * the tasks.
 * 
 * @author diewald
 */
public class SameSilInPartitionConstraintEvaluator extends
		ConstraintEvaluatorBase<Double, SameSilInPartitionConstraint, TaskMappingEncodingBase<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?>> {

	/** Prototype Constructor. */
	public SameSilInPartitionConstraintEvaluator(
			Class<SameSilInPartitionConstraint> explorationTargetType) {
		super(explorationTargetType);
	}

	/** Constructor. */
	public SameSilInPartitionConstraintEvaluator(SameSilInPartitionConstraint explorationTarget) {
		super(explorationTarget);
	}

	/**
	 * Penalty function that ensures only {@link ITaskAdapter} with the same {@link SIL} reside in
	 * the same {@link Partition}. For each {@link Partition}, an exponential penalty is calculated.
	 * The resulting values are summed to provide an estimate for the overall system.
	 */
	@Provides
	@Evaluate(SameSilInPartitionConstraintEvaluator.class)
	public DoubleValue evaluateGoal(TaskPartitionMappingEncoding phenotype) {
		double resVal = 0.0;
		for(Partition partition : phenotype.getResources()) {
			Collection<ITaskAdapter> allocatedTasks = phenotype.getRequesters(partition);
			if(allocatedTasks.isEmpty()) {
				continue;
			}

			int derivationSum = 0;
			int minSIL =
					allocatedTasks.stream().map(t -> t.getRequiredSafetyIntegrityLevel().getValue())
							.reduce((x, y) -> min(x, y)).get();
			for(ITaskAdapter task : allocatedTasks) {
				int currentSilVal = task.getRequiredSafetyIntegrityLevel().getValue();
				int deviation = currentSilVal - minSIL;
				if(deviation > 0) {
					derivationSum += deviation;
				}
			}

			resVal += expm1(derivationSum);
		}
		return new DoubleValue(resVal);
	}

	/** Validates the {@link ExplorationConstraint} versus the penalty value. */
	@Provides
	@Validates
	public Boolean validate(Value<?> value) {
		return value.getDouble() <= 10e-6;
	}
}
