/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform.evaluator;

import static org.fortiss.af3.exploration.alg.feature.platform.PlatformExplorationUtils.updateTargetResourceSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.evaluate.annotations.Evaluate;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.feature.partition.variable.TaskPartitionMappingEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.EvaluatorModule;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEncoding;
import org.fortiss.af3.exploration.alg.module.common.evaluator.mapping.MappingParserObjective;
import org.fortiss.af3.exploration.dseml.model.expression.IExpression;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.opt4j.core.DoubleValue;

import com.google.inject.Inject;
import com.google.inject.Provides;

/**
 * Evaluator for {@link IExpression}s of {@link ExplorationObjective}s. The actual evaluation
 * function resides in the evaluate method.
 * 
 * @author diewald
 */
public class PlatformMappingEvaluatorObjective
		extends EvaluatorModule<DoubleValue, ExplorationTarget<DoubleValue>, Phenotype> {

	/** Constructor. */
	public PlatformMappingEvaluatorObjective(
			Class<? extends ExplorationTarget<DoubleValue>> explorationTargetType) {
		super(explorationTargetType);
	}

	/** Constructor. */
	public PlatformMappingEvaluatorObjective(SMTObjective explorationTarget) {
		super(explorationTarget);
	}

	/** Container for optional dependencies. */
	protected static class OptionalVirtMapping {
		/** Optional virtual mapping (Partitions) */
		@Inject(optional = true)
		@Genotyped
		TaskPartitionMappingEncoding virtMapping;
	}

	/**
	 * Performs an evaluation of the given mapping. The underlying {@link IExpression} is adjusted
	 * to match the target {@link ExecutionUnit} set specified by the
	 * {@link PlatformExecUnitInstanceEncoding}.
	 */
	@Provides
	@Evaluate(PlatformMappingEvaluatorObjective.class)
	public DoubleValue evaluate(
			IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> mapping,
			@Genotyped PlatformExecUnitInstanceEncoding peuiEnc,
			OptionalVirtMapping virtMappingContainer, TaskMapper taskMapper) {
		@SuppressWarnings("unchecked") IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?> virtMapping =
				(IMappingEncoding<ITaskAdapter, IExecutionUnitAdapter, TaskMappingEntry<ITaskAdapter, IExecutionUnitAdapter>, ?>)(IMappingEncoding<?, ?, ?, ?>)virtMappingContainer.virtMapping;

		IExpression modExpr = updateTargetResourceSet(explorationTarget.getExpression(),
				peuiEnc.getContainedInstanceTable(), peuiEnc.getTemplates());

		MappingParserObjective parser =
				new MappingParserObjective(mapping, virtMapping, modExpr, taskMapper);
		Number retVal = null;
		try {
			retVal = parser.parse();
		} catch(Exception e) {
			throw new ExplorationRuntimeException(
					"Failed to parse the expression " + explorationTarget.getName() + ".", e);
		}

		return new DoubleValue(retVal.doubleValue());
	}
}
