/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.platform.transform;

import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.BIDIRECTIONAL;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.INCOMING;
import static org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType.OUTGOING;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getConnectionsBetween;
import static org.fortiss.af3.exploration.util.DSMLModelElementFactory.createSuperSet;
import static org.fortiss.af3.exploration.util.ExplorationModelElementFactory.createContainedByElementSpec;
import static org.fortiss.af3.platform.utils.PlatformModelElementFactory.createTransmissionConnection;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.findElementById;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.decode.Genotyped;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.platform.variable.PlatformExecUnitInstanceEncoding;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.ExplorationTransformationModuleBase;
import org.fortiss.af3.exploration.alg.service.module.base.transformation.Transforms;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.TransmissionConnection;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.Table;
import com.google.common.collect.Table.Cell;
import com.google.inject.Provides;

/**
 * Transforms a {@link PlatformExecUnitInstanceEncoding} into a {@link SuperSet} of
 * {@link ExecutionUnit}s that contains all generated instances.
 * 
 * @author diewald
 */
public class PlatformTransmissionConnectionTransformer
		extends ExplorationTransformationModuleBase<SuperSet<TransmissionConnection>> {

	/**
	 * See {@link PlatformTransmissionConnectionTransformer}.
	 * 
	 * @param execUnits
	 *            Dependency to execute the transformator for execution units first: it instantiates
	 *            the ExecutionUnit elements.
	 * @param peuiEnc
	 *            encoding relating a parent {@link ExecutionUnit} to is contained instances of
	 *            {@link ExecutionUnit}s.
	 * @param pcgEnc
	 *            encoding that defines the connection between platform elements.
	 * @return SuperSet of AF3 {@link ExecutionUnit} instances.
	 */
	@Transforms
	@Provides
	public <N extends EObject, S extends EObject> SuperSet<TransmissionConnection> transform(
			SuperSet<ExecutionUnit> execUnits, @Genotyped PlatformExecUnitInstanceEncoding peuiEnc,
			PlatformCommunicationGraphEncoding pcgEnc) {
		SuperSet<TransmissionConnection> tConnections =
				createSuperSet("TransmissionConnections", TransmissionConnection.class);

		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> pGraph = pcgEnc.getActualGraph();

		for(IExecutionUnitAdapter parentExecUnitAdp : peuiEnc.getAbstractContainerExecUnits()) {
			for(IExecutionUnitAdapter instExecUnitAdp : peuiEnc
					.getChildInstances(parentExecUnitAdp)) {
				// Add the connections from the instance to the TransmissionUnits.
				tConnections.getEntries().addAll(createInstanceToTUnitConnections(pGraph, peuiEnc,
						parentExecUnitAdp, instExecUnitAdp));
			}
		}

		return tConnections;
	}

	/**
	 * Creates incoming {@link IConnection} copies that connect the wrapped {@link IModelElement} of
	 * the given {@code instExecUnitAdp} with the {@link TransmissionUnit}s connected to the
	 * template {@link ExecutionUnit}.
	 */
	private Collection<TransmissionConnection> createInstanceToTUnitConnections(
			DefaultDirectedGraph<IResourceAdapter, DefaultEdge> pGraph,
			PlatformExecUnitInstanceEncoding peuiEnc, IExecutionUnitAdapter parentExecUnitAdp,
			IExecutionUnitAdapter instance) {
		Collection<TransmissionConnection> createdConnections = new ArrayList<>();

		Map<IResourceAdapter, IResourceAdapter> origToInstResourceMap =
				peuiEnc.getContainedInstancesOf(instance);
		Collection<IResourceAdapter> instResourceSet = origToInstResourceMap.values();
		Table<IResourceAdapter, IResourceAdapter, ConnectionType> connectOuterResToInstance =
				HashBasedTable.create();
		for(IResourceAdapter instResource : instResourceSet) {
			for(IResourceAdapter neighbor : Graphs.neighborListOf(pGraph, instResource)) {
				if(!instResourceSet.contains(neighbor)) {
					if(pGraph.getEdge(instResource, neighbor) != null) {
						// Outgoing edge.
						ConnectionType connType = OUTGOING;
						if(connectOuterResToInstance.contains(neighbor, instance)) {
							connType = BIDIRECTIONAL;
						}
						connectOuterResToInstance.put(neighbor, instance, connType);
					}
					if(pGraph.getEdge(neighbor, instResource) != null) {
						// Incoming edge.
						ConnectionType connType = INCOMING;
						if(connectOuterResToInstance.contains(neighbor, instance)) {
							connType = BIDIRECTIONAL;
						}
						connectOuterResToInstance.put(neighbor, instance, connType);
					}
				}
			}
		}

		// Connect the instantiated element with its environment.
		IExecutionUnitAdapter templateAdp = peuiEnc.getTemplateOf(parentExecUnitAdp);
		createdConnections.addAll(
				createConnections(connectOuterResToInstance, templateAdp, parentExecUnitAdp));

		return createdConnections;
	}

	/**
	 * Generates connections for all resource pairs in the given table.
	 * 
	 * @param connectMap
	 *            {@link Table} that associates resource pairs to be connected with the direction of
	 *            their connection.
	 * @param templateAdp
	 *            template of the instantiated {@link IResourceAdapter}.
	 * @param parentExecUnitAdp
	 *            container of the template and instance {@link IResourceAdapter}s.
	 * @return collection of created {@link TransmissionConnection}s.
	 */
	private Collection<TransmissionConnection> createConnections(
			Table<IResourceAdapter, IResourceAdapter, ConnectionType> connectMap,
			IExecutionUnitAdapter templateAdp, IExecutionUnitAdapter parentExecUnitAdp) {
		Collection<TransmissionConnection> createdConnections = new ArrayList<>();

		for(Cell<IResourceAdapter, IResourceAdapter, ConnectionType> connEntry : connectMap
				.cellSet()) {
			IResourceAdapter otherResAdp = connEntry.getRowKey();
			IResourceAdapter instResAdp = connEntry.getColumnKey();
			IPlatformArchitectureElement otherElement =
					otherResAdp.getObject(IPlatformArchitectureElement.class);
			IPlatformArchitectureElement instElement =
					instResAdp.getObject(IPlatformArchitectureElement.class);

			ExecutionUnit parentExecUnit = parentExecUnitAdp.getObject(ExecutionUnit.class);
			createdConnections.addAll(createResourceConnection(parentExecUnit,
					templateAdp.getObject(ExecutionUnit.class), instElement, otherElement,
					connEntry.getValue()));
		}

		return createdConnections;
	}

	/**
	 * Creates a connection between the {@code instExecUnit} and the {@code otherElement}.
	 * 
	 * @param connectionContainer
	 *            container of the created {@link TransmissionConnection}.
	 * @param templateExecUnit
	 *            template {@link ExecutionUnit} from which the {@link TransmissionConnection}s are
	 *            derived.
	 * @param instExecUnit
	 *            instance {@link IPlatformArchitectureElement} that shall be connected.
	 * @param otherElement
	 *            some other element in the same platform layer as the {@code instExecUnit}.
	 * @param connType
	 *            direction of the connection (in/out/bi).
	 * @return collection of created {@link TransmissionConnection}s between the instance and the
	 *         other element.
	 */
	private Collection<TransmissionConnection> createResourceConnection(
			ExecutionUnit connectionContainer, ExecutionUnit templateExecUnit,
			IPlatformArchitectureElement instExecUnit, IPlatformArchitectureElement otherElement,
			ConnectionType connType) {
		Map<TransmissionConnection, ConnectionType> connections = new HashMap<>();

		IPlatformArchitectureElement srcElement = instExecUnit;
		IPlatformArchitectureElement tgtElement = otherElement;

		switch(connType) {
			case OUTGOING:
				connections.putAll(getConnectionsBetween(templateExecUnit, tgtElement));
				break;
			case INCOMING:
				// Swap source and target compared to the outgoing connection.
				srcElement = otherElement;
				tgtElement = instExecUnit;
				connections.putAll(getConnectionsBetween(otherElement, templateExecUnit));
				break;
			default: /* BIDIRECTIONAL */
				connections.putAll(getConnectionsBetween(templateExecUnit, otherElement));
				connections.putAll(getConnectionsBetween(otherElement, templateExecUnit));
		}

		return createConnectionCopies(connectionContainer, connections, srcElement, tgtElement,
				instExecUnit);
	}

	/**
	 * Creates copies of the given {@link IConnection}s and creates new target and source
	 * {@link IConnector}s as needed.
	 */
	private Collection<TransmissionConnection> createConnectionCopies(ExecutionUnit parentExecUnit,
			Map<TransmissionConnection, ConnectionType> connections,
			IHierarchicElement sourceElement, IHierarchicElement targetElement,
			IHierarchicElement instElement) {
		Collection<TransmissionConnection> createdConnections = new ArrayList<>();

		Map<PlatformConnectorUnit, PlatformConnectorUnit> existingToNewSourceConnector =
				new HashMap<>();
		Map<PlatformConnectorUnit, PlatformConnectorUnit> existingToNewTargetConnector =
				new HashMap<>();
		for(Entry<TransmissionConnection, ConnectionType> templateConn : connections.entrySet()) {
			// We may need to swap source and target elements for undirected connections: They have
			// an "artificial" direction in the platform.
			// NOTE: We are assuming only one direction kind (isDirected).
			// if(!isDirected) {
			// if(templateConn.getValue() == INCOMING) {
			// IHierarchicElement tmpElement = sourceElement;
			// sourceElement = targetElement;
			// targetElement = tmpElement;
			// }
			// }

			PlatformConnectorUnit srcConnector = templateConn.getKey().getSource();
			PlatformConnectorUnit newSourceConnector =
					existingToNewSourceConnector.get(srcConnector);
			if(newSourceConnector == null) {
				if(sourceElement == instElement) {
					newSourceConnector =
							(PlatformConnectorUnit)findElementById(srcConnector.getId(),
									sourceElement);
				} else {
					newSourceConnector = copy(srcConnector);
					// This specification indicates where to add the connector in the final model.
					createContainedByElementSpec(newSourceConnector, sourceElement);
				}

				existingToNewSourceConnector.put(srcConnector, newSourceConnector);
			}

			PlatformConnectorUnit tgtConnector = templateConn.getKey().getTarget();
			PlatformConnectorUnit newTargetConnector =
					existingToNewTargetConnector.get(tgtConnector);
			if(newTargetConnector == null) {
				if(targetElement == instElement) {
					newTargetConnector =
							(PlatformConnectorUnit)findElementById(tgtConnector.getId(),
									targetElement);
				} else {
					newTargetConnector = copy(tgtConnector);
					// This specification indicates where to add the connector in the final model.
					createContainedByElementSpec(newTargetConnector, targetElement);
				}

				existingToNewTargetConnector.put(tgtConnector, newTargetConnector);
			}

			// NOTE: We must disable notifications here to avoid creating back references from
			// elements of the snapshot PlatformArchitecture. Thus, the owner, source and target
			// port must be set again by the SuperSet to PlatformArchitecture Transformator.
			TransmissionConnection tConn = createTransmissionConnection();
			createContainedByElementSpec(tConn, parentExecUnit);
			tConn.setSource(newSourceConnector);
			tConn.setTarget(newTargetConnector);

			createdConnections.add(tConn);
		}

		return createdConnections;
	}
}
