/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.taskmapping.problem;

import static org.eclipse.core.runtime.Assert.isNotNull;
import static org.fortiss.af3.exploration.alg.util.GraphUtils.getGraphOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.PhenotypeBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.SignalAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.module.common.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.opt4j.core.Genotype;

/**
 * Encoding (as a {@link Genotype}) which represents the set of deployable Components that shall be
 * mapped to {@link IExecutionUnitAdapter}s by the DSE.
 * 
 * @author diewald
 */
public class TaskGraphEncodingBase<T extends ITaskGraph<S>, S extends ITaskAdapter> extends
		PhenotypeBase implements IExplorationEncoding, Copyable<TaskGraphEncodingBase<T, S>> {
	/** {@link ITaskGraph} that represents the deployable components and their signals. */
	protected Collection<T> taskGraphs;

	/** The Hyperperiod of the application represented by task graphs. */
	protected Double hyperPeriod;

	/**
	 * Collection of {@link ITaskAdapter}s that shall be mapped to deployment targets
	 * ({@link IExecutionUnitAdapter}).
	 */
	protected Collection<ITaskAdapter> taskSet = new HashSet<>();

	/**
	 * Subset of {@link ITaskAdapter}s from the collection of deployable components
	 * that may be replicated.
	 */
	protected Collection<ITaskAdapter> replicableTasks = new HashSet<>();

	/**
	 * Constructor that initializes {@code this} encoding with the collection of
	 * {@link ITaskAdapter}s as they are extracted from the input models.
	 */
	public TaskGraphEncodingBase(Collection<T> taskGraphs, Double hyperPeriod) {
		this.taskGraphs = taskGraphs;
		this.hyperPeriod = hyperPeriod;

		this.taskSet.clear();

		for(ITaskGraph<S> tg : taskGraphs) {
			Collection<? extends ITaskAdapter> tasks = tg.getTasks();
			taskSet.addAll(tasks);
		}
	}

	/**
	 * Copy constructor that takes an existing {@link TaskGraphEncodingBase} as input and
	 * generates an equivalent copy.
	 */
	public TaskGraphEncodingBase(TaskGraphEncodingBase<T, S> encoding) {
		this.hyperPeriod = encoding.getHyperPeriod();
		taskGraphs.addAll(encoding.getTaskGraphs());
		taskSet.addAll(encoding.taskSet);
		replicableTasks.addAll(encoding.replicableTasks);
	}

	/** Returns the collection of {@link ITaskGraph}s composing {@code this} encoding. */
	public Collection<T> getTaskGraphs() {
		return taskGraphs;
	}

	/**
	 * Adds the given {@link ITaskAdapter} to its corresponding {@link ITaskGraph} in {@code this}
	 * encoding.
	 */
	public void addTask(S task) {
		ITaskGraph<S> taskGraph = getGraphOf(task, taskGraphs);
		isNotNull(taskGraph, "The task to be added could be neither found in this encoding, " +
				"nor the input models.");

		taskGraph.addTask(task);
		taskSet.add(task);
		if(task.getMinReplication() != task.getMaxReplication()) {
			replicableTasks.add(task);
		}
	}

	/**
	 * Adds all signals defined in the input models for the given {@link ITaskAdapter} which must be
	 * contained in any of the internal {@link ITaskGraph}s.
	 */
	public void addSignalsOf(S task, ITaskGraph<S> origTaskGraph) {
		// The given task must be in the any of the task graphs!

		T taskGraph = getGraphOf(task, taskGraphs);
		isNotNull(taskGraph);

		Collection<SignalAdapterWeightedEdge> incomingEdges =
				origTaskGraph.getGraph().incomingEdgesOf(task);
		Collection<SignalAdapterWeightedEdge> outgoingEdges =
				origTaskGraph.getGraph().outgoingEdgesOf(task);

		for(SignalAdapterWeightedEdge edge : incomingEdges) {
			S source = taskGraph.getGraph().getEdgeSource(edge);
			taskGraph.getGraph().addEdge(source, task,
					new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
		}

		for(SignalAdapterWeightedEdge edge : outgoingEdges) {
			S target = taskGraph.getGraph().getEdgeTarget(edge);
			taskGraph.getGraph().addEdge(task, target,
					new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
		}
	}

	/**
	 * Adds the signals of a cloned task by querying the existing signals from the clone's original
	 * task.
	 */
	public void addClonedSignalsOf(S originalTask, S clonedTask) {
		T taskGraph = getGraphOf(originalTask, taskGraphs);
		isNotNull(taskGraph);

		Collection<SignalAdapterWeightedEdge> incomingEdges =
				taskGraph.getGraph().incomingEdgesOf(originalTask);
		Collection<SignalAdapterWeightedEdge> outgoingEdges =
				taskGraph.getGraph().outgoingEdgesOf(originalTask);

		for(SignalAdapterWeightedEdge edge : incomingEdges) {
			S source = taskGraph.getGraph().getEdgeSource(edge);
			taskGraph.getGraph().addEdge(source, clonedTask,
					new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
		}

		for(SignalAdapterWeightedEdge edge : outgoingEdges) {
			S target = taskGraph.getGraph().getEdgeTarget(edge);
			taskGraph.getGraph().addEdge(clonedTask, target,
					new SignalAdapterWeightedEdge(edge.getSignalAdapter()));
		}
	}

	/**
	 * Clones the given {@link ITaskAdapter} and adds it to the corresponding task
	 * graph and further data structures.
	 */
	public S cloneTask(S task) {
		// Clone the task and add it to the corresponding task graph.
		ITaskGraph<S> taskGraph = getGraphOf(task, taskGraphs);
		@SuppressWarnings("unchecked") S clonedTask = (S)task.clone();
		taskGraph.addTask(clonedTask);
		addClonedSignalsOf(task, clonedTask);
		// Add the cloned task to the cache sets.
		taskSet.add(clonedTask);
		if(replicableTasks.contains(task)) {
			replicableTasks.add(clonedTask);
		}

		return clonedTask;
	}

	/**
	 * Returns the {@link ITaskGraph} containing the given task or {@code null} if none is found.
	 */
	public T getTaskGraphOf(S task) {
		return getGraphOf(task, taskGraphs);
	}

	/** Whether {@code this} encoding contains the given task. */
	public boolean containsTask(S task) {
		return getGraphOf(task, taskGraphs) != null;
	}

	/**
	 * Removes the given {@link ITaskAdapter} from the {@link ITaskGraph} containing
	 * the particular task and from the caches.
	 * <p>
	 * <i>Side-note:</i> Each task may be located in only one {@link ITaskGraph}.
	 * 
	 * @returns {@code true} iff a containing {@link ITaskGraph} has been found.
	 */
	public boolean removeTask(S task) {
		ITaskGraph<S> taskGraph = getGraphOf(task, taskGraphs);
		if(taskGraph != null) {
			taskGraph.removeTask(task);
			taskSet.remove(task);
			replicableTasks.remove(task);
			return true;
		}
		assert (false); // Never here.
		return false;
	}

	/**
	 * Returns the deployable (software) component that are present in the internal
	 * {@link ITaskGraph}s. May contain "abstract" {@link Component}s.
	 */
	public Set<S> getTasks() {
		// return deployableComponents;
		Set<S> tasks = new HashSet<>();
		for(T tg : taskGraphs) {
			tasks.addAll(tg.getTasks());
		}
		return tasks;
	}

	/** Returns the {@link SignalAdapterWeightedEdge}s of all sub-{@link ITaskGraph}s. */
	public Set<SignalAdapterWeightedEdge> getSignals() {
		Set<SignalAdapterWeightedEdge> signals = new HashSet<>();
		taskGraphs.forEach(tg -> signals.addAll(tg.getSignals()));
		return signals;
	}

	/**
	 * Returns the deployable (software) components that my be
	 * replicated. May contain "abstract" {@link Component}s.
	 */
	public Collection<ITaskAdapter> getReplicableComponents() {
		return replicableTasks;
	}

	/**
	 * Returns the {@link ITaskAdapter} that wraps the given {@link EObject}. Here,
	 * only those {@link ITaskAdapter} are searched which can be actually executed on
	 * {@link ExecutionUnit}s, i.e. "abstract" components are not considered.
	 */
	public ITaskAdapter getReplicableDeployableComponentOf(EObject originalEObject) {
		for(ITaskAdapter deployableComponentAdapter : replicableTasks) {
			if(deployableComponentAdapter.getObject(EObject.class) == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/** Returns the hyper period of the application. */
	public Double getHyperPeriod() {
		return hyperPeriod;
	}

	/** {@inheritDoc} */
	@Override
	public TaskGraphEncodingBase<T, S> copy() {
		return new TaskGraphEncodingBase<T, S>(this);
	}
}
