/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm;

import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ITaskGraph;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEncodingBase;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEntry;

/**
 * Responsible for generating the {@link Message}s that must be sent between different
 * {@link IMappingEntry}s to satisfy the data dependencies that are encoded in the input logical
 * architecture.
 * 
 * @author diewald
 */
public class MessageGenerator<T extends IExecutionUnitAdapter, E extends TaskMappingEntry<ITaskAdapter, T>, M extends TaskMappingEncodingBase<ITaskAdapter, T, E, ?>> {
	/**
	 * Generates messages from the given {#link encoding}. This methods evaluates the {#link
	 * encoding}
	 */
	public Collection<Message> generateMessages(M encoding, TaskGraphEncoding atgEnc) {
		Collection<Message> generatedMessages = new HashSet<Message>();

		for(E entry : encoding.getMappingEntries()) {
			ITaskAdapter senderComponent = entry.getSource();
			if(senderComponent.isInstantiation()) {
				senderComponent = senderComponent.getReplacedComponent();
			}

			ITaskGraph<ITaskAdapter> tg = atgEnc.getTaskGraphOf(senderComponent);

			Collection<ISignalAdapter> emittedMessages = new ArrayList<>();
			tg.getGraph().outgoingEdgesOf(senderComponent)
					.forEach(e -> emittedMessages.add(e.getSignalAdapter()));

			for(ISignalAdapter currentMessageAdapter : emittedMessages) {
				// Create a Collection of dependent MappingEntries, for which messages are generated
				// (if required).
				Collection<E> mappingEntriesOfSuccesors =
						getSuccessorMappingEntries(encoding, currentMessageAdapter);
				Message generatedMessage = generateMessageForSenderEntry(currentMessageAdapter,
						entry, mappingEntriesOfSuccesors);
				if(generatedMessage != null) {
					generatedMessages.add(generatedMessage);
				}
			}
		}

		return generatedMessages;
	}

	/**
	 * Creates a {@link Collection} which contains all {@link IMappingEntry}s that are causally
	 * depend on the given sender {@link ITaskAdapter}.
	 */
	private Collection<E> getSuccessorMappingEntries(M encoding, ISignalAdapter emittedMessage) {
		Collection<E> mappingEntriesOfSuccesors = new HashSet<E>();
		Collection<ITaskAdapter> realSuccessorComponents = new HashSet<>();

		for(ITaskAdapter successorComponent : emittedMessage.getTargets()) {
			if(successorComponent.isInstantiationRequired()) {
				// Since the generated message refer to the components in the system model adapter,
				// we must search the encoding for equivalent components in this special case
				// (instantiation).
				realSuccessorComponents.addAll(successorComponent.getReplacementComponents());
			} else {
				realSuccessorComponents.add(successorComponent);
			}
		}

		for(ITaskAdapter successorComponent : realSuccessorComponents) {
			Collection<ITaskAdapter> encComps = filterSet(encoding.getRequesters(),
					c -> c.getObject(Object.class) == successorComponent.getObject(Object.class));
			for(ITaskAdapter encComp : encComps) {
				mappingEntriesOfSuccesors.addAll(encoding.getMappingEntriesOf(encComp));
			}
		}
		return mappingEntriesOfSuccesors;
	}

	/**
	 * Creates a new Message for the given {#link senderMappingEntry} and its receiver
	 * {@link IMappingEntry}s if the sender and the receiver are not mapped onto the same
	 * {@link IExecutionUnitAdapter}.
	 */
	private Message generateMessageForSenderEntry(ISignalAdapter senderChannel,
			TaskMappingEntry<?, ?> senderMappingEntry, Collection<E> mappingEntriesOfSuccesors) {

		Collection<E> receiverEntries = new HashSet<E>();
		for(E mappingEntryReceiver : mappingEntriesOfSuccesors) {
			if((senderMappingEntry.getTarget() != mappingEntryReceiver.getTarget())) {
				receiverEntries.add(mappingEntryReceiver);
			}
		}

		if(!receiverEntries.isEmpty()) {
			return new Message(senderChannel, senderMappingEntry, receiverEntries);
		}
		return null;
	}
}
