/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.feature.taskmapping.problem.comm;

import static org.fortiss.af3.exploration.alg.util.GraphUtils.getSubGraph;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.feature.platform.PlatformCommunicationGraphEncoding;
import org.jgrapht.GraphPath;
import org.jgrapht.Graphs;
import org.jgrapht.alg.shortestpath.BellmanFordShortestPath;
import org.jgrapht.graph.AbstractGraph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Calculates the shortest path from the sending entity to its receivers on a physical/virtual
 * platform.
 * 
 * @author diewald
 */
public class ShortestPathRouter extends MessageRouter {
	/** Buffers the already calculated shortest paths for the sender {@link IResourceAdapter}. */
	@SuppressWarnings("unused")
	private Map<IResourceAdapter, BellmanFordShortestPath<IResourceAdapter, DefaultEdge>> shortestPathForSender;

	/** Constructs a simple shortest path router for a given exploration run */
	public ShortestPathRouter() {
		// TODO: Replace with more efficient Dijkstra, which needs a modification to contain the
		// shortest paths to all other vertices in the graph.
		shortestPathForSender =
				new ConcurrentHashMap<IResourceAdapter, BellmanFordShortestPath<IResourceAdapter, DefaultEdge>>();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized DefaultDirectedGraph<IResourceAdapter, DefaultEdge>
			calculateMessageRoute(IResourceAdapter source,
					Collection<IResourceAdapter> receiverResources,
					PlatformCommunicationGraphEncoding pcgEncoding) {
		// Generates a graph that contains the message's path to each receiver in the following
		// manner:
		// - Find the shortest path from the sender to each receiver (using BellmanFord)
		// - Construct a graph from the edge list of the shortest path to each receiver
		DefaultDirectedGraph<IResourceAdapter, DefaultEdge> messageGraph =
				new DefaultDirectedGraph<IResourceAdapter, DefaultEdge>(DefaultEdge.class);

		// Check, if the shortest paths have already been calculated for the given sender/source. If
		// not start a new shortest path analysis using BellmanFord.
		// TODO: Disabled caching for now.
		BellmanFordShortestPath<IResourceAdapter, DefaultEdge> shortestPathForCurrentMessage =
				null;
		// try {
		// shortestPathForCurrentMessage = shortestPathForSender.get(source);
		// } catch(IllegalArgumentException e) {
		// pcgEncoding.display();
		// throw e;
		// }

		// if(shortestPathForCurrentMessage == null) {
		shortestPathForCurrentMessage =
				new BellmanFordShortestPath<IResourceAdapter, DefaultEdge>(
						pcgEncoding.getActualGraph());
		// shortestPathForSender.put(source, shortestPathForCurrentMessage);
		// }

		for(IResourceAdapter currentReceiver : receiverResources) {
			GraphPath<IResourceAdapter, DefaultEdge> shortestMessagePath = null;
			try {
				shortestMessagePath =
						shortestPathForCurrentMessage.getPath(source, currentReceiver);
			} catch(IllegalArgumentException e) {
				pcgEncoding.display();
				throw e;
			}

			// No valid path between source and receiver => invalid mapping;
			if(shortestMessagePath == null) {
				return null;
			}
			// Add all vertices and edges.
			Graphs.addAllEdges(messageGraph, pcgEncoding.getActualGraph(),
					shortestMessagePath.getEdgeList());
		}

		return messageGraph;
	}

	/** {@inheritDoc} */
	@Override
	public Map<IResourceAdapter, AbstractGraph<IResourceAdapter, DefaultEdge>>
			getRemainingMessageGraphs(Message message,
					ICommunicationResourceAdapter communicationResource,
					AbstractGraph<IResourceAdapter, DefaultEdge> messageRoute) {
		Map<IResourceAdapter, AbstractGraph<IResourceAdapter, DefaultEdge>> remainingSubGraphs =
				new HashMap<>();

		for(IResourceAdapter receiver : message.getReceiverResources()) {
			AbstractGraph<IResourceAdapter, DefaultEdge> containingGraph =
					getFirst(remainingSubGraphs.values(), g -> g.containsVertex(receiver))
							.orElse(null);
			if(containingGraph == null) {
				containingGraph = new DefaultDirectedGraph<IResourceAdapter, DefaultEdge>(
						DefaultEdge.class);
				getSubGraph(messageRoute, receiver, v -> v == communicationResource,
						containingGraph);
			}

			remainingSubGraphs.put(receiver, containingGraph);
		}

		return remainingSubGraphs;
	}
}
