/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/

package org.fortiss.af3.exploration.alg.feature.taskmapping.variable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.feature.taskmapping.problem.TaskGraphEncodingBase;
import org.fortiss.af3.exploration.alg.module.common.encoding.IMappingEncoding;
import org.fortiss.af3.exploration.alg.module.common.encoding.MappingEncodingBase;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;

/**
 * Base class for encodings of task to execution unit mappings.
 * 
 * @author huang
 * @author diewald
 */
public abstract class TaskMappingEncodingBase<S extends ITaskAdapter, T extends IExecutionUnitAdapter, E extends TaskMappingEntry<S, T>, M extends IMappingEncoding<S, T, E, M>>
		extends MappingEncodingBase<S, T, E, M> {

	/** The input {@link TaskGraphEncodingBase} that contains the tasks and their signals. */
	protected TaskGraphEncodingBase<?, S> tgEncoding;

	/** Ratio between the generation of RoundRobin based and pure random mapping generation. */
	private final double fixToRandomRatio = 0.25;

	/** Constructor that optionally initializes the encoding to a random encoding. */
	protected TaskMappingEncodingBase(Set<T> resources, TaskGraphEncodingBase<?, S> tgEncoding,
			Collection<InternalConstraint<?>> constraints, boolean initialize) {
		super(tgEncoding.getTasks(), resources, constraints);
		this.tgEncoding = tgEncoding;
		if(initialize) {
			createEncoding(tgEncoding.getTasks());
		}
	}

	/** Copy Constructor. */
	protected TaskMappingEncodingBase(TaskMappingEncodingBase<S, T, E, M> encoding) {
		super(encoding);
	}

	/**
	 * Creates {@link TaskMappingEntry}s for the given set of {@link ITaskAdapter}s.
	 * 
	 * @param allocatableRequesters
	 *            Collection of components to generated {@link TaskMappingEntry}s for.
	 */
	protected void createEncoding(Collection<S> allocatableRequesters) {
		double selRandOrFixedInit = rand.nextDouble();
		List<T> execTargets = new ArrayList<>(getResources());
		int roundRobinAllocIdx = rand.nextInt(execTargets.size());

		List<S> compColl = new ArrayList<>(allocatableRequesters);
		if(selRandOrFixedInit < fixToRandomRatio) {
			Collections.shuffle(compColl);
		}

		for(S requester : allocatableRequesters) {
			List<T> validExecTargets = new ArrayList<>();
			validExecTargets.addAll(getValidTargetResources(requester));

			if(selRandOrFixedInit < fixToRandomRatio) {
				createAndAddEntry(requester, null);
			} else {
				for(int repl = 0; repl < requester.getMaxReplication(); repl++) {
					int calcTargetIdx = roundRobinAllocIdx % validExecTargets.size();
					addAllocation(createEntry(requester, validExecTargets.get(calcTargetIdx)));
					roundRobinAllocIdx++;
					if(roundRobinAllocIdx == execTargets.size()) {
						roundRobinAllocIdx = 0;
					}
				}
			}
		}
	}

	/**
	 * Removes the given {@link ITaskAdapter} and, consequently, all associated entries from the
	 * mapping.
	 */
	public void removeTask(S task) {
		allocationTable.row(task).clear();
	}

	/** Assigns deployment targets to a deployable component */
	public void setAssignedResources(S deployableComponent, List<E> deploymentTargets) {
		removeTask(deployableComponent);
		deploymentTargets.forEach(e -> addAllocation(e));
	}

	/** * Returns the size of the encoding. */
	@Override
	public int size() {
		return allocationTable.size();
	}
}
