/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.encoding;

import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.d3seext.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.feature.taskmapping.variable.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.module.common.operator.copy.Copyable;
import org.fortiss.af3.exploration.alg.service.module.IExplorationEncoding;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;
import org.opt4j.core.Genotype;

/**
 * Generalized interface for accessing Request->Resource mappings that are the result of a DSE.
 * 
 * @author diewald
 */
public interface IMappingEncoding<S extends IRequestAdapter, T extends IResourceAdapter, E extends IMappingEntry<S, T>, M extends IMappingEncoding<S, T, E, M>>
		extends Genotype, Phenotype, Copyable<M>, IExplorationEncoding {

	/**
	 * Returns the (sub)set of {@link IRequestAdapter}s to which the given {@code source} may be
	 * allocated.
	 * 
	 * @param source
	 *            {@link IRequestAdapter} to fin valid allocation targets for.
	 * @return Collection of valid target {@link IResourceAdapter}s.
	 */
	List<T> getValidTargetResources(S source);

	/**
	 * Creates an (encoding specific) mapping entry the describes that a given
	 * {@link IResourceAdapter} is mapped to a deployment target.
	 */
	E createEntry(S requester, T resource);

	/**
	 * Creates a {@link TaskMappingEntry} and adds it to {@code this} mapping. The given
	 * {@link IRequestAdapter} and {@link IResourceAdapter} define the source and target for the
	 * created {@link TaskMappingEntry}.
	 * 
	 * @param requester
	 *            Source of the {@link IMappingEntry}.
	 * @param resource
	 *            {@link IRequestAdapter} of the created {@link IMappingEntry}.
	 * @return The created {@link TaskMappingEntry}.
	 */
	E createAndAddEntry(S requester, T resource);

	/**
	 * Constructs an {@IMappingEntry} for the given requester. The deployment targets can be given
	 * as a list, otherwise the deployment targets are taken from the platform.
	 */
	E createEntry(S requester, List<T> deploymentTargets);

	/**
	 * Returns the collection of {@link IMappingEntry}s which represent the (multi-)allocation of
	 * the given {@link IRequestAdapter}s to {@link IResourceAdapter}s.
	 */
	List<E> getMappingEntriesOf(S source);

	/**
	 * Returns the collection of {@link IMappingEntry}s which represent the (multi-)allocation of
	 * the given {@link IRequestAdapter}s to {@link IResourceAdapter}s.
	 */
	List<E> getMappingEntriesOf(T target);

	/**
	 * Returns the collection of {@link IMappingEntry}s contained in {@code this}
	 * {@link IMappingEncoding}.
	 */
	Collection<E> getMappingEntries();

	/** Removes the given {@link IMappingEntry} from the mapping. */
	void removeEntry(E entry);

	/** Resets the target {@link IRequestAdapter} of the given entry. */
	void resetResourceOf(E entry, T resource);

	/** Returns the collection of resource requesters. */
	Set<S> getRequesters();

	/**
	 * Returns the {@link IRequestAdapter}s allocated to the given {@link IResourceAdapter}.
	 * 
	 * @param resource
	 *            Allocation target.
	 * @return Allocated {@link IRequestAdapter}s.
	 */
	List<S> getRequesters(T resource);

	/**
	 * Returns the set of {@link IResourceAdapter}s to which {@link IRequestAdapter}s can be
	 * allocated.
	 */
	Set<T> getResources();

	/**
	 * Returns the allocation of a specific {@link IRequestAdapter} to its {@link IMappingEntry}s.
	 */
	List<T> getResources(S source);

	/**
	 * Updates the internal constraint map to match an updated set of {@link IRequestAdapter}s and
	 * {@link IResourceAdapter}s
	 */
	public void updateValidAllocations(Set<S> requesters, Set<T> resources,
			Collection<InternalConstraint<?>> constraints);

	/**
	 * Validates that the encoding obeys the {@link InternalConstraint}s. Used for debugging. This
	 * method shall only be used in {@link Assert} statements in combination with a prior
	 * DSE_DEBUG_VALIDATE conditional (logic AND).
	 */
	boolean isValidAllocations();
}
