/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.common.input;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.exception.InvalidModelException;
import org.fortiss.af3.exploration.alg.module.InputModule;
import org.fortiss.af3.exploration.alg.module.InputProvider;
import org.fortiss.af3.exploration.dseml.model.expression.SuperSet;
import org.fortiss.af3.task.model.Signal;
import org.fortiss.af3.task.model.TaskInputPort;
import org.fortiss.af3.task.model.TaskOutputPort;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.inject.Provides;
import com.google.inject.Singleton;

/**
 * Creates the internal representation of signals exchanged between tasks.
 * 
 * @author diewald
 */
public class TaskSignalInputModule extends InputModule<Set<ISignalAdapter>> {

	/** See {@link TaskSignalInputModule}. */
	@InputProvider
	@Provides
	@Singleton
	public Set<ISignalAdapter> getSignalAdapters(SuperSet<Signal> signals, TaskMapper tMapper) {
		// TODO: We could also use the signals directly, but that affects the rest of the DSE. Thus
		// this change needs to be considered thoroughly and requires (maybe) large changes.
		Set<ISignalAdapter> taskSignals = new HashSet<>();
		Multimap<TaskOutputPort, TaskInputPort> outputPorts = HashMultimap.create();

		for(Signal sig : signals.getEntries()) {
			outputPorts.put((TaskOutputPort)sig.getSource(), (TaskInputPort)sig.getTarget());
		}
		for(TaskOutputPort outPort : outputPorts.keys()) {
			ITaskAdapter senderTask = tMapper.getAdapter(outPort.getTask());

			Collection<ITaskAdapter> receiverTasks = new HashSet<>();
			for(TaskInputPort targetIn : outputPorts.get(outPort)) {
				receiverTasks.add(tMapper.getAdapter(targetIn.getTask()));
			}
			if(receiverTasks.size() == 0) {
				throw new InvalidModelException(senderTask.getObject(IModelElement.class),
						"Found a signal with a sender Task but no receiver. " +
								"This indicates a faulty input Task Architecture.");
			}

			if(!isPureLoopSignal(senderTask, receiverTasks)) {
				// Remove loops from multicast signals if there are any.
				receiverTasks.remove(senderTask);
				taskSignals.add(new AF3SignalAdapter(outPort, senderTask, outputPorts.get(outPort),
						receiverTasks));
			}
		}

		return taskSignals;
	}

	/** Determines whether the multicast signal only transfers data to the sender (loop). */
	private boolean isPureLoopSignal(ITaskAdapter senderTask,
			Collection<ITaskAdapter> receiverTasks) {
		if(receiverTasks.size() > 1) {
			// A signal with multiple receivers cannot be a "pure" loop: It may send to itself, but
			// this case is handled in the addSignals method.
			return false;
		}

		ITaskAdapter receiver = getFirst(receiverTasks).get();
		if(senderTask == receiver) {
			return true;
		}
		return false;
	}
}
