/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.module.expspectransf.dsmlext;

import static org.fortiss.af3.exploration.alg.util.DesignSpaceExplorationModelElementFactory.createFixedPropertyLiteral;
import static org.fortiss.tooling.base.utils.AnnotationUtils.getAnnotation;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Predicate;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.dseml.model.arithmetic.IArithmeticTermContainer;
import org.fortiss.af3.exploration.dseml.model.arithmetic.ModelElementPropertyLiteral;
import org.fortiss.af3.exploration.dseml.model.expression.IExpression;
import org.fortiss.af3.exploration.lang.ExpressionVisitorBase;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.dsmlext.FixedPropertyLiteral;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.AnnotationValueService;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;

import com.google.common.reflect.TypeToken;
import com.google.inject.Injector;

/**
 * Replaces {@link ModelElementPropertyLiteral}s with {@link FixedPropertyLiteral}s that cache the
 * property values such that they can be accessed more efficiently by the D3SE.
 * 
 * @author diewald
 */
public final class PropertyToFixedValueLiteralConverter extends ExpressionVisitorBase
		implements IExpSpecTransformationModule<ExplorationTarget<?>, ExplorationTarget<?>> {

	/**
	 * Caches the replacement {@link FixedPropertyLiteral}s that are inserted after visiting the
	 * complete {@link IExpression}.
	 */
	Map<ModelElementPropertyLiteral, FixedPropertyLiteral<?>> replaceMap = new HashMap<>();

	/** Boilerplate: "Remembers" the type of the input given as a generic. */
	private final TypeToken<ExplorationTarget<?>> inTypeToken =
			new TypeToken<ExplorationTarget<?>>(getClass()) { // ...
			};

	/** Boilerplate: "Remembers" the type of the output given as a generic. */
	private final TypeToken<ExplorationTarget<?>> outTypeToken =
			new TypeToken<ExplorationTarget<?>>(getClass()) { // ...
			};

	/** {@inheritDoc} */
	@Override
	public boolean outputIsReplacement() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public ExplorationTarget<?> transform(Injector mainInjector, ExplorationTarget<?> input)
			throws ExplorationException {
		if(input.getExpression() == null) {
			// There may exist objectives without DSML expressions. No error.
			return null;
		}

		try {
			input.getExpression().accept(this, Object.class);
		} catch(Exception e) {
			throw(new ExplorationException(
					"Could not transform the DSML expression " + input.toString(), e));
		}

		for(ModelElementPropertyLiteral meLit : getChildrenWithType(input,
				ModelElementPropertyLiteral.class)) {
			FixedPropertyLiteral<?> fixedLit = replaceMap.get(meLit);
			EObject meLitContainer = meLit.eContainer();
			if(!(meLitContainer instanceof IArithmeticTermContainer)) {
				throw new ExplorationException("Unexpected DSML expression: The " +
						ModelElementPropertyLiteral.class.getSimpleName() + " is not contained" +
						"in a " + IArithmeticTermContainer.class.getSimpleName() + ".");
			}
			((IArithmeticTermContainer)meLitContainer).setTerms(fixedLit);
		}

		return input;
	}

	/** {@inheritDoc} */
	@Override
	public Object visit(ModelElementPropertyLiteral mePropLiteral) {
		Map<IModelElement, Object> valueMap = new HashMap<>();
		for(IModelElement me : mePropLiteral.getSetReference().getEntries()) {
			IAnnotatedSpecification spec = getAnnotation(me, mePropLiteral.getSpecificationType());
			AnnotationEntry annEntry = AnnotationValueService.getInstance().getAnnotationEntry(me);
			// Use the spec's class since we need the concrete implementing class for the call
			// below.
			Object annValue = annEntry.getSpecificationValue(spec.getClass());
			valueMap.put(me, annValue);
		}

		FixedPropertyLiteral<?> fixedLit =
				createFixedPropertyLiteral(valueMap, mePropLiteral.getSetReference());
		replaceMap.put(mePropLiteral, fixedLit);
		return null;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<ExplorationTarget<?>> getInputType() {
		return (Class<ExplorationTarget<?>>)inTypeToken.getRawType();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<ExplorationTarget<?>> getOutputType() {
		return (Class<ExplorationTarget<?>>)outTypeToken.getRawType();
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Predicate<ExplorationTarget<?>> getInputFilter() {
		return null;
	}
}
