/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.module.IExpSpecTransformationModule;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.constraints.InternalConstraint;

import com.google.common.collect.BiMap;
import com.google.inject.Injector;

/**
 * This service transforms a collection of input {@link ExplorationTarget}s that can be processed
 * internally by the DSE, e.g. in operators,
 * such that the DSE's efficiency can be increased. It does so by providing registered
 * {@link IExpSpecTransformationModule}s with the generic constraints for which
 * the modules were registered and collecting the transformation results of each module.
 * 
 * @author diewald
 */
public interface IExpSpecTransformationService extends IExplorationService {
	/**
	 * Transforms the {@link ExplorationTarget}s of a DSE run. Examples for this preprocessing
	 * includes a transformation into {@link InternalConstraint}s, or DSML processing.
	 * 
	 * @param mainInjector
	 *            {@link Injector} providing access to the input models.
	 * @param expTargets
	 *            list of input to transformed constraints.
	 * @param origExpTargets
	 *            list of original {@link ExplorationTarget}s that may not be modified but serves as
	 *            a helper to identify replacements.
	 * @param copyRefMap
	 *            Map relating original {@link ExplorationTarget}s and their children with their
	 *            copies that may be modified.
	 * @return Possibly empty collection of transformed {@link ExplorationConstraint}s.
	 * @throws ExplorationException
	 *             If the transformation fails for any reason.
	 */
	<I extends EObject, O extends EObject> List<ExplorationTarget<?>> transform(
			Injector mainInjector, List<ExplorationTarget<?>> expTargets,
			List<ExplorationTarget<?>> origExpTargets, BiMap<EObject, EObject> copyRefMap)
			throws ExplorationException;

	/**
	 * Registers the given module {@code T} (e.g., an {@link IExpSpecTransformationModule}) with
	 * this module/service.
	 * 
	 * @param module
	 *            to register.
	 */
	void registerModule(IExpSpecTransformationModule<? extends EObject, ? extends EObject> module);
}
